#ifndef lint
static char *RCSid = "@(#)$Id: xmexit.c,v 2.0 90/04/09 10:03:42 tot Exp $";
#endif /* not lint */

/* Program to shut down X server. Based on Motif 1.0.x

   Copyright 1990 Front End Oy

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation, and that the name of Front End not be used in advertising
   or publicity pertaining to distribution of the software without specific,
   written prior permission.  Front End makes no representations about the
   suitability of this software for any purpose.  It is provided "as is"
   without express or implied warranty.

   FRONT END DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
   ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
   FRONT END.  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
   ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
   IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
   OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:  Teemu Torma, Front End Oy
            tot@frend.fi

   $Author: tot $
   $Date: 90/04/09 10:03:42 $
   $Source: /vol/src/X/clients/xmexit/xmexit.c,v $
   $Revision: 2.0 $
   $State: Exp $

   $Log:	xmexit.c,v $
   Revision 2.0  90/04/09  10:03:42  tot
   First public release (patchlevel 0).
   
*/

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/MwmUtil.h>
#include <X11/Xos.h>
#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/MessageB.h>     
#include <stdio.h>
#include <signal.h>
#include "xmexit.h"

/* Application resources. */

static XtResource Resources[] = {
  { Nconfirm, CConfirm, XtRBoolean, sizeof (Boolean),
      XtOffset (XMexit, confirm), XtRImmediate, (caddr_t) True },
  { Ndebug, CDebug, XtRBoolean, sizeof (Boolean),
      XtOffset (XMexit, debug), XtRImmediate, (caddr_t) False },
  { Ndelay, CDelay, XtRInt, sizeof (int),
      XtOffset (XMexit, delay), XtRImmediate, (caddr_t) 2 },
  { Ngrab, CGrab, XtRBoolean, sizeof (Boolean),
      XtOffset (XMexit, grab), XtRImmediate, (caddr_t) True },
  { Nlevels, CLevels, XtRInt, sizeof (int),
      XtOffset (XMexit, levels), XtRImmediate, (caddr_t) 0 },
  { NokDefault, COkDefault, XtRBoolean, sizeof (Boolean),
      XtOffset (XMexit, ok_default), XtRImmediate, (caddr_t) False },
  { Ntext, CText, XmRXmString, sizeof (XmString),
      XtOffset (XMexit, text), XtRString, "Exit X-Windows?" },
};

/* Options for us. */

static XrmOptionDescRec Options[] = {
  { "-quiet", "*confirm", XrmoptionNoArg, "False" },
  { "-confirm", "*confirm", XrmoptionNoArg, "True" },
  { "-text", "*text", XrmoptionSepArg },
  { "-ok", "*okDefault", XrmoptionNoArg, "True" },
  { "-cancel", "*okDefault", XrmoptionNoArg, "False" },
  { "-levels", "*levels", XrmoptionSepArg },
  { "-grab", "*grab", XrmoptionNoArg, "True" },
  { "-nograb", "*grab", XrmoptionNoArg, "False" },
  { "-delay", "*delay", XrmoptionSepArg },
  { "-debug", "*debug", XrmoptionNoArg, "True" },
};

/* Atoms we need to use. */

static Atom WmState;
static Atom WmProtocols;
static Atom WmDeleteWindow;
static Atom MotifMwmInfo;

/* If set, we are debugging.
   Just show actions we should make. */

static Boolean Debug = False;


/* Show usage of this program. */

static void
Usage ()
{
  fputs ("\
usage:\n\
        xmexit [-options]\n\
\n\
where options include X toolkit options and\n\
    -quiet                  Do not show confirm dialog\n\
    -confirm                Do show confirm dialog (default)\n\
    -text <string>          Text to show in dialog\n\
    -ok                     Default action is OK\n\
    -cancel                 Default action in Cancel (deafult)\n\
    -grab                   Grab the server (default)\n\
    -nograb                 Do not grab the server\n\
    -levels <num>           Look applications only NUM levels downs\n\
    -delay <num>            Wait NUM seconds before force kill\n\
", stderr);
  exit (1);
}


/* Replacement for XGetWMProtocols.
   This is needed since Motif stores WM_PROTOCOLS as WM_PROTOCOLS type instead
   of ATOM. */

static Status
XGetWMProtocols (display, window, protocols, num_protocols)
     Display *display;
     Window window;
     Atom **protocols;
     int *num_protocols;
{
  Atom *data = (Atom *) 0;
  Atom actual_type;
  int actual_format;
  unsigned long leftover, nitems;

  /* Get the property. */

  if (XGetWindowProperty (display, window, WmProtocols, 0, 1000000, False,
			  AnyPropertyType, &actual_type, &actual_format,
			  &nitems, &leftover, (unsigned char **) &data)
      != Success)
    return False;

  /* Type may be ATOM or WM_PROTOCOLS. */

  if ((actual_type != XA_ATOM && actual_type != WmProtocols)
      || actual_format != 32)
    {
      if (data)
	XFree ((char *) data);
      return False;
    }

  *protocols = data;
  *num_protocols = (int) nitems;
  return True;
}


/* Delete WINDOW of DISPLAY by using WM_DELETE_WINDOW protocol. */

static void
DeleteWindow (display, window)
     Display *display;
     Window window;
{
  XClientMessageEvent event;

  event.display = display;
  event.window = window;
  event.type = ClientMessage;
  event.send_event = True;
  event.message_type = WmProtocols;
  event.format = 32;
  event.data.l[0] = WmDeleteWindow;

  XSendEvent (display, window, False, 0, (XEvent *) &event);
}


/* See if WINDOW has property PROP.
   Return True if it has, False if doesn't */

static Boolean
PropertyExist (display, window, prop)
     Display *display;
     Window window;
     Atom prop;
{
  int num_props, count;
  Atom *props;

  props = XListProperties (display, window, &num_props);
  if (props)
    {
      for (count = 0; count < num_props; count ++)
	if (props[count] == prop)
	  {
	    XFree ((char *) props);
	    return True;
	  }
      XFree ((char *) props);
    }
  return False;
}


/* Kill application for WINDOW.
   If FORCE, always do XKillClient, otherwise try to use WM_DELETE_WINDOW
   protocol if window has one. */

static void
KillApplication (xmexit, window, force)
     XMexit xmexit;
     Window window;
     Boolean force;
{
  Display *display = xmexit->display;
  Atom *protocols;
  int num_protocols;

  /* If this is our window, skip. */

  if (window == xmexit->top
      || (xmexit->dialog && window == xmexit->dialog))
    {
      if (Debug)
	printf ("Skipping window %#8x as own window\n", window);
      return;
    }

  /* If window has WM_PROCOTOLS property, see if it contains WM_DELETE_WINDOW
     property. */

  if (! force && PropertyExist (display, window, WmProtocols)
      && XGetWMProtocols (display, window, &protocols, &num_protocols))
    {
      int count;
      Boolean found = False;

      for (count = 0; count < num_protocols; count ++)
	if (protocols[count] == WmDeleteWindow)
	  {
	    found = True;
	    break;
	  }

      if (found)
	{
	  if (Debug)
	    printf ("Kill window %#8x using WM_DELETE_WINDOW\n", window);
	  else
	    DeleteWindow (display, window);
	}
      else
	{
	  if (Debug)
	    printf ("Kill window %#8x using XKillClient\n", window);
	  else
	    XKillClient (display, window);
	}

      XFree ((char *) protocols);
    }
  else if (Debug)
    printf ("Kill window %#8x using XKillClient\n", window);
  else
    XKillClient (display, window);
}      


/* Add WINDOW to xmexit's window list. */

static void
AddWindowToList (xmexit, window)
     XMexit xmexit;
     Window window;
{
  if (! xmexit->window_list)
    {
      xmexit->window_list = (Window *) XtMalloc (100 * sizeof (Window));
      xmexit->window_list_size = 100;
      xmexit->last_window = 0;
    }
  else if (xmexit->last_window >= xmexit->window_list_size)
    {
      xmexit->window_list_size += 100;
      xmexit->window_list = (Window *)
	XtRealloc ((char *) xmexit->window_list,
		   xmexit->window_list_size * sizeof (Window));
    }

  xmexit->window_list[xmexit->last_window ++] = window;
}


/* Our own error handler.
   If error is BadWindow in certain of our requests, ignore it. */

static int
ErrorHandler (display, error)
     Display *display;
     XErrorEvent *error;
{
  switch (error->request_code)
    {
    case X_QueryTree:
    case X_ListProperties:
    case X_SendEvent:
    case X_KillClient:
    case X_GetProperty:

      /* Ignore BadWindow and BadValue errors since we will get them.
	 It's because killing windows usually cause destroying other windows
	 also since normally clients have more than just one window. */

      if (error->error_code == BadValue || error->error_code == BadWindow)
	return 0;
      break;
    }

  return _XDefaultError (display, error);
}


/* Read windows to xmexit's window list below ROOT window.
   LEVEL is current level we are at.
   If ALL, read all windows, regardless of LEVEL and whether WM_STATE property
   exists (except our windows, of cource). */

static void
ReadWindows (xmexit, root, level, all)
     XMexit xmexit;
     Window root;
     int level;
     Boolean all;
{
  Display *display = xmexit->display;
  Window dummy, *kids;
  unsigned int num_kids;
  int count;

  /* If we are below maximum level, nothing to do. */

  if (! all && xmexit->levels && level > xmexit->levels)
    return;

  if (! XQueryTree (display, root, &dummy, &dummy, &kids, &num_kids))
    return;
	    
  /* Add kids with WM_STATE property to window list. They are potential X
     clients.
     If not possible X client, recurse tree below it. */

  for (count = 0; count < num_kids; count ++)
    if (kids[count] != xmexit->top
	&& (! xmexit->dialog || kids[count] != xmexit->dialog))
      {
	if (all)
	  {
	    AddWindowToList (xmexit, kids[count]);
	    ReadWindows (xmexit, kids[count], level + 1, all);
	  }
	else if (PropertyExist (display, kids[count], WmState))
	  AddWindowToList (xmexit, kids[count]);
	else
	  ReadWindows (xmexit, kids[count], level + 1, all);
      }
  
  XFree ((char *) kids);
}


/* Read MWM's window.
   MWM stores it to root window's property _MOTIF_WM_INFO. */

static void
ReadMwmWindow (xmexit, root)
     XMexit xmexit;
     Window root;
{
  PropMwmInfo *property = (PropMwmInfo *) 0;
  Atom actual_type;
  int actual_format;
  unsigned long nitems, leftover;

  if (XGetWindowProperty (xmexit->display, root, MotifMwmInfo, 0,
			  PROP_MWM_INFO_ELEMENTS, False, MotifMwmInfo,
			  &actual_type, &actual_format, &nitems, &leftover,
			  (unsigned char **) &property) != Success)
    return;
  
  if (actual_type == MotifMwmInfo && nitems == PROP_MWM_INFO_ELEMENTS)
    AddWindowToList (xmexit, property->wmWindow);

  if (property)
    XFree ((char *) property);
}


/* Kill all the clients from the X server. */

static void
KillClients (xmexit)
     XMexit xmexit;
{
  Display *display = xmexit->display;
  int screen, count;

  /* Intern atoms we need. */

  WmState = XmInternAtom (display, "WM_STATE", False);
  WmProtocols = XmInternAtom (display, "WM_PROTOCOLS", False);
  WmDeleteWindow = XmInternAtom (display, "WM_DELETE_WINDOW", False);
  MotifMwmInfo = XmInternAtom (display, _XA_MWM_INFO, False);

  /* Initialize window list variables in the xmexit structure. */

  xmexit->window_list = (Window *) 0;
  xmexit->window_list_size = 0;
  xmexit->last_window = 0;

  /* Ignore hangup signals.
     If we are invoked under xterm or emacs, they will kill us when we kill
     them. I wonder whether we should ignore also SIGTERM's.. */

  signal (SIGHUP, SIG_IGN);

  /* Set our own error handler that ignores BadWindow errors. */

  XSetErrorHandler (ErrorHandler);

  /* Read in top level windows in all screens. */

  for (screen = 0; screen < ScreenCount (display); screen ++)
    {
      ReadWindows (xmexit, RootWindow (display, screen), 1, False);

      /* Try to get MWM's window from the root window.
	 MWM is silly since it's window does not have WM_STATE set... */

      ReadMwmWindow (xmexit, RootWindow (display, screen));
    }

  /* Kill the windows in the window list. */

  for (count = 0; count < xmexit->last_window; count ++)
    KillApplication (xmexit, xmexit->window_list[count], False);

  /* Remove our windows from the display.
     Just for the visual effect. */

  XUnmapWindow (display, xmexit->top);
  if (xmexit->dialog)
    XUnmapWindow (display, xmexit->dialog);

  XSync (display, False);

  /* If we've got delay, wait that many seconds before doing force kill.
     This time, we read all the windows and do XKillClient(3X11) to them. */

  if (xmexit->delay > 0)
    {
      sleep (xmexit->delay);

      /* Remove all windows from the window list. */

      xmexit->last_window = 0;

      /* Read all windows into the memory and kill them. */

      for (screen = 0; screen < ScreenCount (display); screen ++)
	{
	  ReadWindows (xmexit, RootWindow (display, screen), 1, True);
	  ReadMwmWindow (xmexit, RootWindow (display, screen));
	}
      for (count = 0; count < xmexit->last_window; count ++)
	KillApplication (xmexit, xmexit->window_list[count], True);
      XSync (display, False);
    }
}


/* Confirm callback which never returns.
   This is called when use confirms  killing all applications. */

static void
ConfirmCallback (w, closure, data)
     Widget w;
     caddr_t closure, data;
{
  XMexit xmexit = (XMexit) closure;

  KillClients (xmexit);
  exit (0);
}


/* Cancel callback.
   Just exit the application. */

static void
CancelCallback (w, closure, data)
     Widget w;
     caddr_t closure, data;
{
  exit (1);
}


/* Popup confirm dialog.
   PARENT is parent widget for us, XMEXIT is application data structure. */

static void
PopupConfirmDialog (parent, xmexit)
     Widget parent;
     XMexit xmexit;
{
  Widget shell, box;
  Arg args[20];
  Cardinal num;
  Widget children[20];
  Dimension width, height;
  Position x, y;
  Display *display = XtDisplay (parent);
  int screen = XScreenNumberOfScreen (XtScreen (parent));

  num = 0;
  XtSetArg (args[num], XmNallowShellResize, True); num ++;
  shell = XtCreatePopupShell ("confirmDialogShell", xmDialogShellWidgetClass,
			      parent, args, num);

  num = 0;
  XtSetArg (args[num], XmNdefaultPosition, False); num ++;
  XtSetArg (args[num], XmNmessageString, xmexit->text); num ++;
  XtSetArg (args[num], XmNdialogType, XmDIALOG_QUESTION); num ++;
  XtSetArg (args[num], XmNdefaultButtonType, xmexit->ok_default ?
	    XmDIALOG_OK_BUTTON : XmDIALOG_CANCEL_BUTTON); num ++; 
  XtSetArg (args[num], XmNdialogStyle, xmexit->grab
	    ? XmDIALOG_SYSTEM_MODAL : XmDIALOG_MODELESS); num ++;
  box = XtCreateWidget ("confirmDialog", xmMessageBoxWidgetClass, shell,
			args, num);

  /* Unmanage some unnecessary widgets. */

  num = 0;
  children[num ++] = XmMessageBoxGetChild (box, XmDIALOG_HELP_BUTTON);
  children[num ++] = XmMessageBoxGetChild (box, XmDIALOG_SEPARATOR);
  XtUnmanageChildren (children, num);

  /* Add callbacks for different actions.
     We shouldn't have any other possibilities than OK or Cancel. */

  XtAddCallback (box, XmNcancelCallback, CancelCallback, (caddr_t) xmexit);
  XtAddCallback (box, XmNokCallback, ConfirmCallback, (caddr_t) xmexit);

  /* Destroy callback if user closes dialog.
     Just call confirm callback as we would have cancelled the operation. */

  XtAddCallback (box, XtNdestroyCallback, CancelCallback, (caddr_t) xmexit);

  /* Realize the shell and center it to the window. */

  XtRealizeWidget (shell);

  num = 0;
  XtSetArg(args[num], XmNheight, &height); num ++;
  XtSetArg(args[num], XmNwidth, &width); num ++;
  XtGetValues (box, args, num);

  x = DisplayWidth (display, screen) / 2 - (int) width / 2;
  y = DisplayHeight (display, screen) / 2 - (int) height / 2;

  num = 0;
  XtSetArg (args[num], XmNx, x); num ++;
  XtSetArg (args[num], XmNy, y); num ++;
  XtSetValues (box, args, num);

  /* Save dialog shell's window to xmexit structure so that killing will know
     to pass it. */

  xmexit->dialog = XtWindow (shell);

  /* Pop this shell up. */

  XtManageChild (box);
}


int
main (argc, argv)
     int argc;
     char **argv;
{
  Widget top;
  XMexitRec xmexit;

  /* Initialize X toolkit and make top level shell.
     Move this widget to somewhere where it's not visible. */

  top = XtInitialize (*argv, "XMexit", Options, XtNumber (Options),
		      &argc, argv);
  if (argc > 1)
    Usage ();

  XtSetMappedWhenManaged (top, False);
  XtRealizeWidget (top);

  /* Get application resources for us.
     We'll need first to register Xm converters first. */

  XmRegisterConverters ();
  XtGetApplicationResources (top, (caddr_t) &xmexit, Resources,
			     XtNumber (Resources), (ArgList) 0, 0);
  Debug = xmexit.debug;

  /* Set our top level windows. */

  xmexit.display = XtDisplay (top);
  xmexit.top = XtWindow (top);
  xmexit.dialog = 0;

  /* If we need to be confirmed, pop up confirm shell.
     Otherwise just kill the clients. */

  if (xmexit.confirm)
    {
      PopupConfirmDialog (top, &xmexit);
      XtMainLoop ();
    }
  else
    KillClients (&xmexit);

  exit (0);
}
